/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002 - 2004
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 * 
 * $Id: MainFrame.java,v 1.6 2004/11/25 11:56:13 dwagelaa Exp $
 * $Log: MainFrame.java,v $
 * Revision 1.6  2004/11/25 11:56:13  dwagelaa
 * Version 0.1.22 release
 *
 * Revision 1.5  2004/08/12 10:01:06  dwagelaa
 * * updated library support and better integration with CoCompose tool
 * * updated MainFrame interface
 *
 * Revision 1.3  2004/05/04 13:45:12  dwagelaa
 * updated library submitted at meeting with Willem on 4-may-2004
 *
 * Revision 1.2  2004/04/27 12:52:24  dwagelaa
 * added preliminary CoCompose Library support.
 *
 * New features: fixed-pitch font, support for CRQL queries + ranking,
 * more visual real estate for query manipulation system, added tooltips,
 * Library can now encode/decode and store newlines (and hence
 * indentation!) of queries, query save will sort list but also 
 * keep current selection (instead of item 0), confirmation
 * for initial migration and indexing, and for deletion of query
 */

package EDU.utwente.CoCompose.gui;

import java.awt.*;
import java.awt.event.*;

import javax.swing.*;
import javax.swing.border.*;
import java.io.*;
import java.util.*;
import org.xml.sax.SAXException;
import EDU.utwente.CoCompose.lang.*;
import EDU.utwente.CoCompose.util.*;
import EDU.utwente.CoCompose.util.Library;
import EDU.utwente.CoCompose.xml.SAXWriter;

/**
 * Main window containing the main menu and composite repository. This window
 * starts up all model editor windows and any other windows.
 *
 * @author Dennis Wagelaar
 * @author Willem Hajenius
 * @version 0.1.22, 12 August 2004
 */

public class MainFrame extends JFrame
  implements RepositoryContainer, ErrorReporter
{
  /**
   * Utility class for display in query result list
   * @author Willem Hajenius
   */
  class ListData
  {
    public String uri = "", name = "(anonymous)";
    public ImageIcon icon = null;
  }

  /**
   * Auxiliary class to display icons alongside text in a list box
   * @author Willem Hajenius
   */
  class ListBoxRenderer extends JLabel implements ListCellRenderer
  {
    public ListBoxRenderer()
    {
      setOpaque(true);
      setHorizontalAlignment(LEFT);
      setVerticalAlignment(CENTER);
    }

    public Component getListCellRendererComponent(JList list,
      Object value, int index, boolean isSelected, boolean cellHasFocus)
    {
      if (isSelected)
      {
        setBackground(list.getSelectionBackground());
        setForeground(list.getSelectionForeground());
      }
      else
      {
        setBackground(list.getBackground());
        setForeground(list.getForeground());
      }

      ListData contents = (ListData) value;
      setText (contents.name);
      setToolTipText (contents.uri);
      setIcon (contents.icon);
      return this;
    }
  }

  class Migrator extends Thread
  {
    java.util.List files;
    String srcdir, dstdir;

    public Migrator (File srcdir, File dstdir, java.util.List files)
    {
      this.srcdir = srcdir.getAbsolutePath();
      if (this.srcdir.endsWith("/") || this.srcdir.endsWith("\\"))
          this.srcdir = this.srcdir.substring(0, this.srcdir.length() - 1);
      this.dstdir = dstdir.getAbsolutePath();
      if (this.dstdir.endsWith("/") || this.dstdir.endsWith("\\"))
          this.dstdir = this.dstdir.substring(0, this.dstdir.length() - 1);
      this.files = files;
    }

    /**
     * Start the thread and the migration procedure
     */
    public void run()
    {
      // Migrate each of these components
      ListIterator li = files.listIterator();
      ProgressMonitor pm = new ProgressMonitor(null, "Migrating component:", "?", 0, files.size());
      pm.setMillisToDecideToPopup(100);
      pm.setMillisToPopup(200);
      int n = 0;
      while (li.hasNext())
      {
        File f = (File) li.next();
        pm.setNote (f.getPath());
        migrateComponent (f);
        pm.setProgress (++n);
      }
      pm.close();
    }

    
    /**
     * Migrate a single component; the original filename gets
     * the suffix ".old".
     * @param f The component's file
     */
    private void migrateComponent (File f)
    {
      try
      {
        if (!f.exists()) return;
        String filepath = f.getAbsolutePath();

        // Construct the appropriate new file name (= dstdir + relative path to srcdir)
        String newpath = dstdir + filepath.substring(srcdir.length(), filepath.length());       
        
        // Load the CompositeType with the old XML parser
        EDU.utwente.CoCompose.lang.Composite composite = new CompositeType (filepath);

        // Create a new model, and add the composite
        String name = f.getName();
        if (name.endsWith(".xml")) name = name.substring(0, name.length() - 4);
        Model model = new Model ();    
        model.insertNode (composite);

        File newfile = new File (newpath);
        newfile.getParentFile().mkdirs();
        newfile.createNewFile();
        SAXWriter xmlWriter = new SAXWriter();
        xmlWriter.setOutputStream (new FileOutputStream (newfile));
        model.toXML (xmlWriter, false, newfile.getAbsolutePath());

      }
      catch (SAXException saxException)
      {
        System.out.println("An error occurred during creation of the XML document");
      }
      catch (Exception exc)
      {
        System.out.println("There was an error parsing the XML file: " + exc.getMessage());
      }
    }
  }


  //////////////////////////////////////////////////////////////////////////////
  // Attributes
  //////////////////////////////////////////////////////////////////////////////

  private static final String DEFAULT_QUERY = "SELECT l.* FROM library l " + 
      "WHERE (l.type = 'concept' OR l.type = 'composite-type')";
  
  private int modelFrameCount_ = 0;
  private javax.swing.Timer statusTimeout_;
  ImageIcon appIcon = new ImageIcon("images/cocompose_icon.gif");
  ImageIcon aboutIcon = new ImageIcon("images/concept.gif");
  ImageIcon newIcon = new ImageIcon("images/new.gif");
  ImageIcon openIcon = new ImageIcon("images/open.gif");
  ImageIcon helpIcon = new ImageIcon("images/help.gif");
  ImageIcon exitIcon = new ImageIcon("images/door.gif");
  ImageIcon refreshIcon = new ImageIcon("images/refresh.gif");
  JMenuBar menuBar1 = new JMenuBar();
  JMenu menuFile = new JMenu();
  JMenuItem menuFileExit = new JMenuItem();
  JMenu menuHelp = new JMenu();
  JMenuItem menuHelpAbout = new JMenuItem();
  JLabel statusBar = new JLabel();
  BorderLayout borderLayout1 = new BorderLayout();
  CompoundBorder titledBorder1;
  JMenuItem menuFileNew = new JMenuItem();
  JMenuItem menuFileOpen = new JMenuItem();
  JPanel jPanel2 = new JPanel();
  JButton newBtn = new JButton();
  JButton openBtn = new JButton();
  GridLayout gridLayout1 = new GridLayout();
  JMenuItem menuAPIHelp = new JMenuItem();
  JMenuItem menuToolHelp = new JMenuItem();
  JMenuItem MenuReadme = new JMenuItem();

  //Extra items to interface with Library
  Library library = Library.instance();

  JPanel         panelLibrary = new JPanel();
  CompoundBorder borderLibrary;

  JPanel         queryPanel = new JPanel();
  JPanel         savedQueryPanel = new JPanel();
  JComboBox      comboSavedQueries = new JComboBox();
  JPanel         panelSavedQueryButtons = new JPanel();
  GridLayout     layoutSavedQueryButtons = new GridLayout();
  JButton        buttonSaveQuery = new JButton(new ImageIcon("images/iconSaveQuery.gif"));
  JButton        buttonLoadQuery = new JButton(new ImageIcon("images/iconLoadQuery.gif"));
  JButton        buttonDeleteQuery = new JButton(new ImageIcon("images/iconDeleteQuery.gif"));

  JTextArea      textQuery = new JTextArea(9, 30);
  JScrollPane    scrollQuery;

  JPanel         panelQueryButtons = new JPanel();
  GridLayout     layoutQueryButtons = new GridLayout();
  JButton        buttonClearQuery = new JButton("Clear", new ImageIcon("images/iconClearQuery.gif"));
  JButton        buttonExecuteQuery = new JButton("Execute", new ImageIcon("images/iconExecuteQuery.gif"));

  DefaultListModel  listmodelQueryResults = new DefaultListModel();
  JList             listQueryResults = new JList(listmodelQueryResults);
  JScrollPane       scrollQueryResults;

  ImageIcon      conceptIcon = new ImageIcon("images/concept.gif");
  ImageIcon      compositeIcon = new ImageIcon("images/composite.gif");
  ImageIcon      solutionIcon = new ImageIcon("images/model.gif");
  ImageIcon      unknownIcon = new ImageIcon("images/iconUnknown.gif");

  JMenuItem      menuReindexLibrary = new JMenuItem();
  JMenuItem      menuRefreshLibrary = new JMenuItem();
  JMenuItem      menuMigrateComponents = new JMenuItem();

  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs the MainFrame.
   */
  public MainFrame()
  {
    enableEvents(AWTEvent.WINDOW_EVENT_MASK);
    try
    {
      jbInit();
      mInit();
    }
    catch(Exception e)
    {
      e.printStackTrace();
    }
  } //MainFrame

  //////////////////////////////////////////////////////////////////////////////
  // Private methods
  //////////////////////////////////////////////////////////////////////////////

  /* JBuilder Component initialization. */
  private void jbInit() throws Exception
  {
    final MainFrame mainframe = this;
    textQuery.setFont (new Font ("Monospaced", Font.PLAIN, 12));
    titledBorder1 = BorderFactory.createCompoundBorder(
      new TitledBorder(BorderFactory.createEtchedBorder(
        Color.white, new java.awt.Color(148, 145, 140)), "Main menu"),
          BorderFactory.createEmptyBorder(0,1,1,1));
    this.getContentPane().setLayout(borderLayout1);
    this.setSize(new Dimension(211, 670));
    this.setTitle("CoCompose");
    statusBar.setBorder(BorderFactory.createEtchedBorder());
    statusBar.setText("Ready");
    menuFile.setIcon(newIcon);
    menuFile.setText("File");
    menuFile.setMnemonic(KeyEvent.VK_F);
    menuFileExit.setIcon(exitIcon);
    menuFileExit.setText("Exit");
    menuFileExit.setAccelerator(javax.swing.KeyStroke.getKeyStroke(115, java.awt.event.KeyEvent.ALT_MASK, false));
    menuFileExit.addActionListener(new java.awt.event.ActionListener()
    {

      public void actionPerformed(ActionEvent e)
      {
        fileExit_actionPerformed(e);
      }
    });
    menuHelp.setIcon(helpIcon);
    menuHelp.setText("Help");
    menuHelp.setMnemonic(KeyEvent.VK_H);
    menuHelpAbout.setIcon(aboutIcon);
    menuHelpAbout.setText("About");
    menuHelpAbout.addActionListener(new java.awt.event.ActionListener()
    {

      public void actionPerformed(ActionEvent e)
      {
        helpAbout_actionPerformed(e);
      }
    });
    menuFileOpen.setIcon(openIcon);
    menuFileOpen.setText("Open model");
    menuFileOpen.setAccelerator(javax.swing.KeyStroke.getKeyStroke(79, java.awt.event.KeyEvent.CTRL_MASK, false));
    menuFileOpen.addActionListener(new java.awt.event.ActionListener()
    {

      public void actionPerformed(ActionEvent e)
      {
        menuFileOpen_actionPerformed(e);
      }
    });
    menuFileNew.setIcon(newIcon);
    menuFileNew.setText("New model");
    menuFileNew.setAccelerator(javax.swing.KeyStroke.getKeyStroke(78, java.awt.event.KeyEvent.CTRL_MASK, false));
    menuFileNew.addActionListener(new java.awt.event.ActionListener()
    {

      public void actionPerformed(ActionEvent e)
      {
        menuFileNew_actionPerformed(e);
      }
    });
    newBtn.setHorizontalAlignment(SwingConstants.LEFT);
    newBtn.setIcon(newIcon);
    newBtn.setText("New");
    newBtn.addActionListener(new java.awt.event.ActionListener()
    {

      public void actionPerformed(ActionEvent e)
      {
        menuFileNew_actionPerformed(e);
      }
    });
    newBtn.setToolTipText("Create a new model");
    openBtn.setIcon(openIcon);
    openBtn.setText("Open");
    openBtn.addActionListener(new java.awt.event.ActionListener()
    {

      public void actionPerformed(ActionEvent e)
      {
        menuFileOpen_actionPerformed(e);
      }
    });
    openBtn.setToolTipText("Open a model");
    openBtn.setHorizontalAlignment(SwingConstants.LEFT);
    jPanel2.setLayout(gridLayout1);
    gridLayout1.setColumns(2);
    gridLayout1.setRows(1);
    jPanel2.setBorder(titledBorder1);

    menuReindexLibrary.setIcon(refreshIcon);
    menuReindexLibrary.setText("Reindex library");
    menuReindexLibrary.setAccelerator(KeyStroke.getKeyStroke("ctrl R"));
    menuReindexLibrary.addActionListener(new java.awt.event.ActionListener()
    {
      public void actionPerformed(ActionEvent e)
      {
        indexComponents();
      }
    });
    menuRefreshLibrary.setIcon(refreshIcon);
    menuRefreshLibrary.setText("Refresh library");
    menuRefreshLibrary.setAccelerator(KeyStroke.getKeyStroke("F5"));
    menuRefreshLibrary.addActionListener(new java.awt.event.ActionListener()
    {
      public void actionPerformed(ActionEvent e)
      {
        library.refreshComponents(false);
      }
    });
    menuMigrateComponents.setText("Migrate old components");
    menuMigrateComponents.addActionListener(new java.awt.event.ActionListener()
    {
      public void actionPerformed(ActionEvent e)
      {
        migrateComponents();
      }
    });

    menuAPIHelp.setIcon(helpIcon);
    menuAPIHelp.setText("API Help");
    menuAPIHelp.addActionListener(new java.awt.event.ActionListener()
    {
      public void actionPerformed(ActionEvent e)
      {
        menuAPIHelp_actionPerformed(e);
      }
    });
    menuToolHelp.setIcon(helpIcon);
    menuToolHelp.setText("Tool Help");
    menuToolHelp.addActionListener(new java.awt.event.ActionListener()
    {
      public void actionPerformed(ActionEvent e)
      {
        menuToolHelp_actionPerformed(e);
      }
    });
    MenuReadme.setIcon(helpIcon);
    MenuReadme.setText("Readme");
    MenuReadme.addActionListener(new java.awt.event.ActionListener()
    {
      public void actionPerformed(ActionEvent e)
      {
        menuReadme_actionPerformed(e);
      }
    });
    menuFile.add(menuFileNew);
    menuFile.add(menuFileOpen);
    menuFile.addSeparator();

    menuFile.add(menuReindexLibrary);
    menuFile.add(menuRefreshLibrary);
    menuFile.add(menuMigrateComponents);

    menuFile.addSeparator();
    menuFile.add(menuFileExit);
    menuHelp.add(MenuReadme);
    menuHelp.add(menuToolHelp);
    menuHelp.add(menuAPIHelp);
    menuHelp.addSeparator();
    menuHelp.add(menuHelpAbout);
    menuBar1.add(menuFile);
    menuBar1.add(menuHelp);
    this.setJMenuBar(menuBar1);

    this.getContentPane().add(panelLibrary, BorderLayout.CENTER);

    this.getContentPane().add(statusBar, BorderLayout.SOUTH);
    this.getContentPane().add(jPanel2, BorderLayout.NORTH);
    jPanel2.add(newBtn, null);
    jPanel2.add(openBtn, null);
    statusTimeout_ = new javax.swing.Timer(2000, new ActionListener()
    {

      public void actionPerformed(ActionEvent e)
      {
        statusBar_timeout(e);
      }
    });
    statusTimeout_.setRepeats(false);

    //Added items to interface with Library
    Insets spacer = new Insets(2, 3, 2, 3);
    borderLibrary = BorderFactory.createCompoundBorder(
      new TitledBorder(BorderFactory.createEtchedBorder(
        Color.white, new java.awt.Color(148, 145, 140)), "Component Library"),
          BorderFactory.createEmptyBorder(0,1,1,1));
    panelLibrary.setBorder(borderLibrary);
    panelLibrary.setLayout(new BorderLayout());
    queryPanel.setLayout(new BorderLayout());
    savedQueryPanel.setLayout(new BorderLayout());
    ((BorderLayout) panelLibrary.getLayout()).setVgap(2);
    ((BorderLayout) queryPanel.getLayout()).setVgap(2);
    ((BorderLayout) savedQueryPanel.getLayout()).setVgap(2);

    // Saved query manipulation
    refillSavedQueriesBox();
    comboSavedQueries.setToolTipText ("Query names");
    comboSavedQueries.setEditable(true);
    comboSavedQueries.setMaximumRowCount(20);
    comboSavedQueries.setAlignmentX(0.5f);
    /*comboSavedQueries.addItemListener (new java.awt.event.ItemListener()
    {
      public void itemStateChanged (ItemEvent e)
      {
        loadQuery();
        executeQuery();
      }
    });*/
    savedQueryPanel.add(comboSavedQueries, BorderLayout.CENTER);

    buttonSaveQuery.setToolTipText("Save the current query");
    buttonSaveQuery.setMargin(spacer);
    buttonSaveQuery.setMnemonic(KeyEvent.VK_S);
    buttonSaveQuery.setHorizontalAlignment(SwingConstants.CENTER);
    buttonSaveQuery.addActionListener(new java.awt.event.ActionListener()
    {
      public void actionPerformed(ActionEvent e)
      {
        String queryName = (String) comboSavedQueries.getSelectedItem();
        queryName = queryName.trim();
        if (queryName.equals("default"))
        {
          mainframe.reportError(new CoComposeException("Cannot overwrite the default query"));
          return;
        }
        if (queryName.length() == 0)
          setStatus ("Can't save query; no name given!");
        else
        {
          String queryContents = (String) textQuery.getText();
          library.saveQuery(queryName, queryContents);
          refillSavedQueriesBox();
          comboSavedQueries.setSelectedItem (queryName);
          setStatus("Saved this query as '" + queryName + "'");
        }
      }
    });

    buttonDeleteQuery.setToolTipText("Delete the selected query");
    buttonDeleteQuery.setMargin(spacer);
    buttonDeleteQuery.setMnemonic(KeyEvent.VK_D);
    buttonDeleteQuery.setHorizontalAlignment(SwingConstants.CENTER);
    buttonDeleteQuery.addActionListener(new java.awt.event.ActionListener()
    {
      public void actionPerformed(ActionEvent e)
      {
        String queryName = (String) comboSavedQueries.getSelectedItem();
        if (queryName.equals("default"))
        {
          mainframe.reportError(new CoComposeException("Cannot delete the default query"));
          return;
        }
        if (JOptionPane.showConfirmDialog (mainframe,
          "Delete query '" + queryName + "'?",
          "Confirm deletion", JOptionPane.YES_NO_OPTION,
          JOptionPane.QUESTION_MESSAGE) == JOptionPane.YES_OPTION)
        {
          library.deleteQuery(queryName);
          refillSavedQueriesBox();
          setStatus("Query '" + queryName + "' deleted");
        }
      }
    });

    buttonLoadQuery.setToolTipText("Load the selected query into the query editor");
    buttonLoadQuery.setMargin(spacer);
    buttonLoadQuery.setMnemonic(KeyEvent.VK_L);
    buttonLoadQuery.setHorizontalAlignment(SwingConstants.CENTER);
    buttonLoadQuery.addActionListener(new java.awt.event.ActionListener()
    {
      public void actionPerformed(ActionEvent e)
      {
        loadQuery();
      }
    });

    layoutSavedQueryButtons.setRows(1);
    layoutSavedQueryButtons.setColumns(3);
    panelSavedQueryButtons.setLayout(layoutSavedQueryButtons);
    panelSavedQueryButtons.add(buttonSaveQuery);
    panelSavedQueryButtons.add(buttonDeleteQuery);
    panelSavedQueryButtons.add(buttonLoadQuery);

    savedQueryPanel.add(panelSavedQueryButtons, BorderLayout.SOUTH);
    queryPanel.add(savedQueryPanel, BorderLayout.NORTH);
    
    // Query manipulation
    textQuery.setLineWrap(true);
    textQuery.setWrapStyleWord(true);
    scrollQuery = new JScrollPane(textQuery);
    scrollQuery.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
    queryPanel.add(scrollQuery, BorderLayout.CENTER);

    buttonClearQuery.setToolTipText("Clear the query editor");
    buttonClearQuery.setMargin(spacer);
    buttonClearQuery.setMnemonic(KeyEvent.VK_C);
    buttonClearQuery.setHorizontalAlignment(SwingConstants.LEFT);
    buttonClearQuery.addActionListener(new java.awt.event.ActionListener()
    {
      public void actionPerformed(ActionEvent e)
      {
        textQuery.setText(null);
      }
    });

    buttonExecuteQuery.setToolTipText("Execute the current query");
    buttonExecuteQuery.setMargin(spacer);
    buttonExecuteQuery.setMnemonic(KeyEvent.VK_X);
    buttonExecuteQuery.setHorizontalAlignment(SwingConstants.LEFT);
    buttonExecuteQuery.addActionListener(new java.awt.event.ActionListener()
    {
      public void actionPerformed(ActionEvent e)
      {
        executeQuery();
      }
    });

    layoutQueryButtons.setRows(1);
    layoutQueryButtons.setColumns(2);
    panelQueryButtons.setLayout(layoutQueryButtons);
    panelQueryButtons.add(buttonClearQuery);
    panelQueryButtons.add(buttonExecuteQuery);
    queryPanel.add(panelQueryButtons, BorderLayout.SOUTH);
    panelLibrary.add(queryPanel, BorderLayout.NORTH);

    // Query results
    listQueryResults.setToolTipText ("Query results");
    listQueryResults.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
    listQueryResults.setVisibleRowCount(12);
    listQueryResults.setCellRenderer(new ListBoxRenderer());    
    listQueryResults.addMouseListener(new MouseAdapter()
    {
      public void mouseClicked(MouseEvent e)
      {
        if (e.getClickCount() == 2)
        {
          int index = listQueryResults.locationToIndex(e.getPoint());
          if (index != -1) openQueryResult(index);
        }
      }
    });
    listQueryResults.addKeyListener(new KeyAdapter()
    {
      public void keyPressed(KeyEvent k)
      {
        if (k.getKeyCode() == KeyEvent.VK_ENTER)
        {
          int index = listQueryResults.getSelectedIndex();
          if (index != -1) openQueryResult(index);
        }
      }
    });
     
    scrollQueryResults = new JScrollPane(listQueryResults);
    scrollQueryResults.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
    panelLibrary.add(scrollQueryResults, BorderLayout.CENTER);

    this.addComponentListener(new ComponentListener()
    {
      public void componentResized(ComponentEvent e) {}
      public void componentMoved(ComponentEvent e) {}
      public void componentHidden(ComponentEvent e) {}
      public void componentShown(ComponentEvent e)
      {
        onMainFrameShown(e);
      }
    });
  } //jbInit

  /* Manual initialization. */
  private void mInit()
  {
    // set window icon
    setIconImage(appIcon.getImage());
    // set window height to 80% or minimum height
    Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
    Dimension size = getSize();
    size.height = Math.max(size.height, (int) (screenSize.height * 0.8));
    setSize(size);
    // set default query
    textQuery.setText(DEFAULT_QUERY);
  } //mInit


  /**
   * Open a selected component in a new window
   * @param index The index in the result list
   */
  void openQueryResult(int index)
  {
    try
    {
      setStatus("Loading model...");
      ListData listrow = (ListData) listmodelQueryResults.get (index);
      Model model = new Model(listrow.uri, listrow.name);
      ModelFrame modelFrame = buildModelFrame(model);
      modelFrame.setTitle(listrow.name); // set window title
      setStatus("Model loaded");
    }
    catch (Exception exc)
    {
      reportError(new CoComposeException("There was an error parsing the XML file: " +
        exc.getMessage()));
      setStatus("Error loading model");
    }
  }


  /**
   * Reload the Saved Queries combo box with the values from the database.
   * Entries are sorted alphabetically.
   */
  void refillSavedQueriesBox()
  {
    comboSavedQueries.removeAllItems();
    comboSavedQueries.addItem("default");
    Vector savedQueryContents = library.sqlQuerySimple("SELECT name FROM SavedQueries", "name");
    java.util.Collections.sort(savedQueryContents);
    Iterator it = savedQueryContents.iterator();
    while (it.hasNext()) comboSavedQueries.addItem(it.next());
  }


  /**
   * When main frame is loaded (and consequently the statusbar becomes available),
   * check whether to reindex the components or not.
   * @param e The ComponentEvent responsible for this method's execution
   */
  private void onMainFrameShown(ComponentEvent e)
  {
    if (library.tableSize ("Library") == 0)
    {
      if (JOptionPane.showConfirmDialog (this,
        "Since CoCompose version 0.1.20 composite types are stored in a " +
        "new file format.\nIf you have composites type prior to this " +
        "version, you can migrate them now.\nIf you choose No, you may " +
        "still migrate them later by using the option in the File menu.\n\n" +
        "Start the migration procedure now?",
        "Start migration procedure?", JOptionPane.YES_NO_OPTION,
        JOptionPane.QUESTION_MESSAGE) == JOptionPane.YES_OPTION)
        migrateComponents();

      if (JOptionPane.showConfirmDialog (this,
        "The new CoCompose repository features a query system to " +
        "retrieve components \nwith certain properties. Before this " +
        "system can be used, the components must be\n'indexed' to extract " +
        "the relevant information.\n\n" +
        "Start the indexing procedure now?",
        "Start indexing procedure?", JOptionPane.YES_NO_OPTION,
        JOptionPane.QUESTION_MESSAGE) == JOptionPane.YES_OPTION)
        indexComponents();
    }
    executeQuery();
  }


  /**
   * (Re)Indexes all components
   */
  private void indexComponents()
  {
    setStatus("Started indexing.");
    library.indexComponents();
    setStatus("Completed indexing.");
  }


  /* Bring up a model editor window. */
  private ModelFrame buildModelFrame()
  {
    return buildModelFrame(new Model());
  } //buildModelFrame()

  /* Bring up a model editor window. */
  private ModelFrame buildModelFrame(Model model_in)
  {
    Dimension frameSize = getSize();
    ModelFrame model_window = new ModelFrame(model_in);

    // Pad window to main tool window
    int cascade = modelFrameCount_ % 8;
    model_window.setLocation(frameSize.width + cascade * 20, cascade * 20);
    modelFrameCount_++;

    model_window.setTitle("CoCompose Design Model " + modelFrameCount_);
    model_window.setRepositoryContainer(this);

    model_window.pack();

    // Set window size
    Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
    Dimension modelFrameSize = model_window.getSize();
    modelFrameSize.height = Math.max(modelFrameSize.height, frameSize.height);
    modelFrameSize.width = Math.min(modelFrameSize.height + 200,
      screenSize.width - frameSize.width);
    model_window.setSize(modelFrameSize);

    model_window.show();

    return model_window;
  } //buildModelFrame(Model model_in)

  /* File | Exit action performed. */
  void fileExit_actionPerformed(ActionEvent e)
  {
    library.shutdown();
    System.exit(0);
  } //fileExit_actionPerformed

  /* Help | About action performed. */
  void helpAbout_actionPerformed(ActionEvent e)
  {
    setStatus("About");
    MainFrame_AboutBox dlg = new MainFrame_AboutBox(this);
    Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
    Dimension dlgSize = dlg.getPreferredSize();
    if (dlgSize.height > screenSize.height)
      dlgSize.height = screenSize.height;
    if (dlgSize.width > screenSize.width)
      dlgSize.width = screenSize.width;
    dlg.setLocation((screenSize.width - dlgSize.width) / 2, (screenSize.height - dlgSize.height) / 2);
    dlg.setModal(true);
    dlg.show();
  } //helpAbout_actionPerformed

  /* Help | Tool Help action. Opens CoCompose tool help. */
  void menuToolHelp_actionPerformed(ActionEvent e)
  {
    File helpFile;
    TextViewDialog viewDlg;

    try
    {
      helpFile = new File("./help/index.html");
      viewDlg = new TextViewDialog(this, "CoCompose Tool Help", false);
      viewDlg.setPage(helpFile.toURL());
      viewDlg.show();
    }
    catch (Exception exc)
    {
      reportError(new CoComposeException("Could not open tool help file: " + exc.getMessage()));
    }
  } //menuToolHelp_actionPerformed

  /* Help | API Help action. Opens CoCompose API help. */
  void menuAPIHelp_actionPerformed(ActionEvent e)
  {
    File helpFile;
    TextViewDialog viewDlg;

    try
    {
      helpFile = new File("./doc/index.html");
      viewDlg = new TextViewDialog(this, "CoCompose API Help", false);
      viewDlg.setPage(helpFile.toURL());
      viewDlg.show();
    }
    catch (Exception exc)
    { // javadoc API not available, display alternative page
      try
      {
        helpFile = new File("./doc/no-index.html");
        viewDlg = new TextViewDialog(this, "CoCompose API Help", true);
        viewDlg.setPage(helpFile.toURL());
        viewDlg.show();
      }
      catch (Exception exc2)
      {
        reportError(new CoComposeException("Could not open API help file: " + exc2.getMessage()));
      }
    }
  } //menuAPIHelp_actionPerformed

  /* Menu | Readme action. Opens readme file. */
  void menuReadme_actionPerformed(ActionEvent e)
  {
    File helpFile;
    TextViewDialog viewDlg;

    try
    {
      helpFile = new File("./Readme.html");
      viewDlg = new TextViewDialog(this, "CoCompose Readme", false);
      viewDlg.setPage(helpFile.toURL());
      viewDlg.show();
    }
    catch (Exception exc)
    {
      reportError(new CoComposeException("Could not open readme file: " + exc.getMessage()));
    }
  } //menuReadme_actionPerformed

  /* Create new model frame. */
  void menuFileNew_actionPerformed(ActionEvent e)
  {
    setStatus("Creating new model...");
    buildModelFrame();
    setStatus("New model created");
  } //menuFileNew_actionPerformed

  /* Resets statusbar text. */
  void statusBar_timeout(ActionEvent e)
  {
    statusBar.setText("Ready");
  } //statusBar_timeout

  /* Opens a model and creates a new model frame that contains it. */
  void menuFileOpen_actionPerformed(ActionEvent e)
  {
    JFileChooser openDlg = new JFileChooser("./models");
    DirFileFilter filter = new DirFileFilter(".xml");
    try
    {
      // filter files on .xml
      openDlg.setFileFilter(filter);
      openDlg.setDialogTitle("Open model");
      if (openDlg.showOpenDialog(this) != JFileChooser.APPROVE_OPTION)
        return;
      File file = openDlg.getSelectedFile();
      // set status
      setStatus("Loading model...");
      // filename chosen - open this file
      String uri = file.getAbsolutePath();
      try
      {
        String name = file.getName();
        if (name.endsWith(".xml"))
          name = name.substring(0, name.length() - 4);
        Model model = new Model(uri, name);
        ModelFrame modelFrame = buildModelFrame(model);
        // set window title
        modelFrame.setTitle(name);
        // set status
        setStatus("Model loaded");
      }
      catch (Exception exc)
      {
        reportError(new CoComposeException("There was an error parsing the XML file: " +
          exc.getMessage()));
        // set status
        setStatus("Error loading model");
      }
    }
    catch (Exception exc)
    {
      reportError(new CoComposeException("It appears your VM does not allow file loading: " +
        exc.getMessage()));
    }
  } //menuFileOpen_actionPerformed

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Overridden event processing so we can exit on System Close.
   *
   * @param e The window event.
   */
  protected void processWindowEvent(WindowEvent e)
  {
    super.processWindowEvent(e);
    if(e.getID() == WindowEvent.WINDOW_CLOSING)
    {
      fileExit_actionPerformed(null);
    }
  } //processWindowEvent

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////


  /**
   * Migrate the components (actually only composites) from the previous
   * versions of CoCompose to the current unifying file format.
   */
  public void migrateComponents()
  {
    try
    {
      JFileChooser source = new JFileChooser(".");
      source.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
      source.setDialogTitle("Select source directory");
      source.setApproveButtonText("Select");
      if (source.showOpenDialog(this) != JFileChooser.APPROVE_OPTION) return;
      File srcdir = source.getSelectedFile();

      JFileChooser dest = new JFileChooser(".");
      dest.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
      dest.setDialogTitle("Select destination directory");
      dest.setApproveButtonText("Select");
      if (dest.showOpenDialog(this) != JFileChooser.APPROVE_OPTION) return;
      File dstdir = dest.getSelectedFile();

      setStatus ("Started migration.");
      LinkedList files = new LinkedList();
      Library.getFileTree (srcdir, ".xml", files); 
      Migrator mig = new Migrator (srcdir, dstdir, files);
      mig.start();
      setStatus ("Completed migration.");
    }
    catch (Exception exc)
    {
      reportError(new CoComposeException("It appears your VM does not allow file loading: " +
        exc.getMessage()));
    }
  }


  private void loadQuery()
  {
    String queryName = (String) comboSavedQueries.getSelectedItem();
    String queryContents;
    if (queryName.equals("default"))
      queryContents = DEFAULT_QUERY;
    else
      queryContents = library.loadQuery(queryName);
    textQuery.setText(queryContents);
    setStatus("Query '" + queryName + "' loaded");
  }


  /* Executes the query entered in the text box and lists the results. */
  private void executeQuery()
  {
    setStatus("Executing query...");
    String query = textQuery.getText();
    Vector results = library.query(query);
    listmodelQueryResults.removeAllElements();
    if (results != null && results.size() > 0)
    {
      int uriIndex = -1, nameIndex = -1, typeIndex = -1;
      Vector header = (Vector) results.get (0);
      uriIndex = header.indexOf ("URI");
      nameIndex = header.indexOf ("NAME");
      typeIndex = header.indexOf ("TYPE");

      for (int i = 1; i < results.size(); i++)
      {
        ListData listdata = new ListData();

        if (nameIndex != -1)
          listdata.name = (String) ((Vector) results.get(i)).get(nameIndex);

        if (uriIndex != -1)
          listdata.uri = (String) ((Vector) results.get(i)).get(uriIndex);

        if (typeIndex != -1)
        {
          String typename = (String) ((Vector) results.get(i)).get(typeIndex);
          if (typename.equalsIgnoreCase ("concept"))
            listdata.icon = conceptIcon;
          else if(typename.equalsIgnoreCase ("composite-type"))
            listdata.icon = compositeIcon;
          else if(typename.equalsIgnoreCase ("solution-pattern"))
            listdata.icon = solutionIcon;
          else listdata.icon = unknownIcon;
        }
        listmodelQueryResults.addElement(listdata);
      }
    }
    setStatus (listmodelQueryResults.size() + " result" +
     (listmodelQueryResults.size() == 1 ? "." : "s."));
  } //executeQuery
  

  /**
   * Construct repository tree from repository directory. Call this method
   * whenever something changes in the composite repository.
   */
  public void buildRepository()
  {
    indexComponents();
  } //buildRepository()


  /**
   * Get the component currently selected in the result list
   * @return A subclass of AbstractConcept
   */
  public AbstractConcept getSelected()
  {
    try
    {
      int index = listQueryResults.getSelectedIndex();
      ListData listrow = (ListData) listmodelQueryResults.get (index);
      Model model = new Model(listrow.uri, listrow.name);
      
      if (model.getFreeNodeCount() == 1) // It's a Concept or a Composite 
	  { 
	    Concept concept = model.firstConcept(); 
	    if (concept != null) 
	    { 
	      Concept copy = new Concept(); 
	      copy.copyAttributes(concept); 
	      return copy; 
	    } 
	    else 
	    { 
	      EDU.utwente.CoCompose.lang.Composite composite = 
	      model.firstComposite(); 
	      if (composite != null) 
	      { 
	        EDU.utwente.CoCompose.lang.Composite copy = 
	          new EDU.utwente.CoCompose.lang.Composite(); 
	        copy.copyAttributes(composite); 
	        return copy; 
	      } 
	      else return null; // Some solitary element of unknown type 
	    } 
	  } 
	  else return null; // Model has multiple elements    
    }
    catch (Exception exc)
    {
      reportError(new CoComposeException("There was an error parsing the XML file: " +
        exc.getMessage()));
      return null;
    }
  }


  /**
   * Reports an error.
   *
   * @param error - the error to report
   */
  public void reportError(CoComposeException error)
  {
    AbstractConcept concept = error.getConcept();
    String type = error.getType();
    String message = error.getMessage();
    if (concept != null)
      concept.setFillColor(Color.red);
    MessageDialog errDlg = new MessageDialog(this, type, true);
    // switch type
    if (type.equals("Error"))
      errDlg.setIcon(MessageDialog.ICON_ERROR);
    else if (type.equals("Warning"))
      errDlg.setIcon(MessageDialog.ICON_WARNING);
    else if (type.equals("Info"))
      errDlg.setIcon(MessageDialog.ICON_INFO);
    errDlg.setMessage(message);
    errDlg.show();
  } //reportError

  /**
   * Accepts generic notification message.
   *
   * @param sender The message sender.
   */
  public void notify(Object sender)
  {
  } //notify

  /**
   * Sets status bar text and updates window.
   *
   * @param status The status text.
   */
  public void setStatus(String status)
  {
    statusBar.setText(status);
    // force repaint
    Graphics g = getGraphics();
    paint(g);
    g.dispose();
    // restart status timer
    statusTimeout_.restart();
  } //setStatus
} //class MainFrame
