/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package EDU.utwente.CoCompose.lang;

import java.awt.*;
import EDU.auburn.VGJ.graph.Edge;
import EDU.auburn.VGJ.graph.Node;
import EDU.auburn.VGJ.graph.GMLobject;
import EDU.auburn.VGJ.util.DPoint3;
import EDU.auburn.VGJ.util.Matrix44;
import EDU.utwente.CoCompose.util.CoComposeException;

/**
 * Represents a link between nodes. Added some extra functionality
 * to the Edge class to use it with CoCompose models.
 *
 * @author Dennis Wagelaar
 * @version 0.0.26, 24 June 2002
 */

public class Link extends Edge
{
  //////////////////////////////////////////////////////////////////////////////
  // Constants
  //////////////////////////////////////////////////////////////////////////////

  private static final Color DEFAULT_SELECT_COLOR = Color.black;
  /**
   * The default color for drawing edges.
   */
  public static final Color DEFAULT_EDGE_COLOR   = new Color(153, 0, 51);

  //////////////////////////////////////////////////////////////////////////////
  // Attributes
  //////////////////////////////////////////////////////////////////////////////

  private Color edgeColor_ = DEFAULT_EDGE_COLOR;
  private Color selectColor_ = DEFAULT_SELECT_COLOR;

  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a Link.
   *
   * @param tail The tail node in the link.
   * @param head The head node in the link.
   * @param points The intermediate points for drawing.
   */
  public Link(Node tail, Node head, DPoint3[] points)
  {
    super(tail, head, points, false);
  } //Link(Node tail, Node head, DPoint3[] points)

  /**
   * Constructs a Link.
   *
   * @param tail The tail node in the link.
   * @param head The head node in the link.
   * @param from The edge to copy the attributes from.
   */
  public Link(Node tail, Node head, Edge from)
  {
    super(tail, head, from);
  } //Link(Node tail, Node head, Edge from)

  /**
   * Constructs a Link.
   *
   * @param tail The tail node in the link.
   * @param head The head node in the link.
   * @param gml The GML object to read the attributes from.
   */
  public Link(Node tail, Node head, GMLobject gml)
  {
    super(tail, head, gml);
  } //Link(Node tail, Node head, GMLobject gml)

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Sets the selection artifact color for drawing.
   *
   * @param selectColor The color.
   */
  public void setSelectColor(Color selectColor)
  {
    selectColor_ = selectColor;
  } //setSelectColor

  /**
   * Gets the selection artifact color for drawing.
   *
   * @return The select color for drawing.
   */
  public Color getSelectColor()
  {
    return selectColor_;
  } //getSelectColor

  /**
   * Sets the edge color for drawing.
   *
   * @param edgeColor The color.
   */
  public void setEdgeColor(Color edgeColor)
  {
    edgeColor_ = edgeColor;
  } //setEdgeColor

  /**
   * Gets the edge color for drawing.
   *
   * @return The edge color for drawing.
   */
  public Color getEdgeColor()
  {
    return edgeColor_;
  } //getEdgeColor

  /**
   * Draws the link onto the graphics context.
   *
   * @param graphics The graphics context.
   * @param transform The coordinate transformation matrix.
   * @param inplane If true, assumes drawing in XY plane.
   * @param quality The drawing quality (> 1 is high quality).
   */
  public synchronized void draw(Graphics graphics, Matrix44 transform,
    boolean inplane, int quality)
  {
    double scale = transform.scale;
    int npoints = points_.length;
    // get intermediate points, if any
    DPoint3 p1to, p2to;
    if (npoints == 0)
    {
      p1to = head_.getPosition3();
      p2to = tail_.getPosition3();
    }
    else
    {
      p1to = points_[0];
      p2to = points_[npoints - 1];
    }
    // get endpoints
    DPoint3 p1 = tail_.intersectWithLineTo(p1to, inplane, quality);
    DPoint3 p2 = head_.intersectWithLineTo(p2to, inplane, quality);
    p1.transform(transform);
    p2.transform(transform);
    // Self edge with no intermediate points.
    if (head_ == tail_ && npoints == 0)
    {
      p2.x = p1.x + 1;
      p2.y = p1.y;
    }
    // draw link
    graphics.setColor(getEdgeColor());
    if (npoints == 0)
    {
      graphics.drawLine((int) p1.x, (int) p1.y, (int) p2.x, (int) p2.y);
    }
    else
    {
      DPoint3 point = new DPoint3(points_[0]);
      point.transform(transform);
      graphics.drawLine((int) p1.x, (int) p1.y, (int) point.x, (int) point.y);
      DPoint3 oldpoint = new DPoint3();
      for (int i = 1; i < npoints; i++)
      {
        oldpoint.move(point);
        point.move(points_[i]);
        point.transform(transform);
        graphics.drawLine((int) oldpoint.x, (int) oldpoint.y, (int) point.x, (int) point.y);
      }
      graphics.drawLine((int)point.x, (int)point.y, (int)p2.x, (int)p2.y);
    }
    // Draw selection handles.
    if (selected)
    {
      graphics.setColor(getSelectColor());
      graphics.drawRect((int) p1.x - 2, (int) p1.y - 2, 4, 4);
      if (head_ != tail_)  // Not a self-edge.
        graphics.drawRect((int)p2.x - 2, (int)p2.y - 2, 4, 4);
      for (int pointindex = 0; pointindex < npoints; pointindex++)
      {
        p2.move(points_[pointindex]);
        p2.transform(transform);
        graphics.drawRect((int) p2.x - 2, (int) p2.y - 2, 4, 4);
      }
      graphics.setColor(Color.white);
      graphics.drawRect((int) p1.x - 1, (int) p1.y - 1, 2, 2);
      if (head_ != tail_)  // Not a self-edge.
        graphics.drawRect((int) p2.x - 1, (int) p2.y - 1, 2, 2);
      for (int pointindex = 0; pointindex < npoints; pointindex++)
      {
        p2.move(points_[pointindex]);
        p2.transform(transform);
        graphics.drawRect((int) p2.x - 1, (int) p2.y - 1, 2, 2);
      }
    }
  } //draw

  /**
   * Gets the Postscript representation for this link.
   *
   * @param transform The coordinate transformation matrix (3D -> 2D).
   * @param inplane If true, the drawing is done in the XY plane.
   */
  public String toPS(Matrix44 transform, boolean inplane)
  {
    String result = new String();
    Color color = getEdgeColor();
    double r = color.getRed() / 255.0;
    double g = color.getGreen() / 255.0;
    double b = color.getBlue() / 255.0;

    result += PSnum_(r) + PSnum_(g) + PSnum_(b) + " setrgbcolor\n";
    result += super.toPS(transform, inplane, false);
    return result;
  } //toPS

  /**
   * Accepts a ModelVisitor.
   *
   * @param visitor The visitor object to accept.
   * @throws CoComposeException if visiting operation failed.
   */
  public void accept(ModelVisitor visitor) throws CoComposeException
  {
    visitor.visitLink(this);
  } //accept
} //class Link
