/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002, 2003
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package EDU.utwente.CoCompose.translate;

import java.io.File;
import java.util.*;
import EDU.utwente.CoCompose.lang.*;
import EDU.utwente.CoCompose.util.*;

/**
 * Retrieves the flattening options for a composite.
 *
 * @author Dennis Wagelaar
 * @version 0.1.17, 19 November 2003
 */

public class FlatteningOptionsVisitor extends ModelVisitor implements Cloneable
{
  //////////////////////////////////////////////////////////////////////////////
  // Attributes
  //////////////////////////////////////////////////////////////////////////////

  // The target language for flattening
  private String language_ = new String();
  // The flattening options
  private Collection options_ = new Vector();
  // cache of validated solution patterns (per model)
  private AppliedPatterns appliedPatterns_ = null;
  /**
   * The set of solution patterns used in the flattening path.
   */
  protected Set patternsInPath = new HashSet();

  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a FlatteningOptionsVisitor.
   */
  public FlatteningOptionsVisitor()
  {
  } //FlatteningOptionsVisitor

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Gets the target language to use for flattening.
   * @return the target language.
   */
  protected String getTargetLanguage()
  {
    return language_;
  } //getTargetLanguage

  /**
   * Adds an option to the collection of flattening options.
   * @param option - the option to add.
   */
  protected void addOption(SolutionPattern option)
  {
    options_.add(option);
  } //addOption

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Resets this visitor to the state it was in just after creation.
   */
  public void reset()
  {
    language_ = new String();
    options_ = new Vector();
    patternsInPath = new HashSet();
    appliedPatterns_ = null;
  } //reset

  /**
   * Returns a copy of this object.
   * @return a copy of this object.
   * @throws CloneNotSupportedException.
   */
  public Object clone() throws CloneNotSupportedException
  {
    FlatteningOptionsVisitor copy = (FlatteningOptionsVisitor) super.clone();
    copy.patternsInPath = new HashSet(patternsInPath);
    copy.options_ = new Vector(); // don't copy options
    return copy;
  } //clone

  /**
   * Sets the target language to use for flattening.
   * @param language - the target language.
   */
  public void setTargetLanguage(String language)
  {
    language_ = language;
  } //setTargetLanguage

  /**
   * Sets the appliedPatterns property for caching of validated solution
   * patterns.
   * @param patterns - the AppliedPatterns object.
   */
  public void setAppliedPatterns(AppliedPatterns patterns)
  {
    appliedPatterns_ = patterns;
  } //setAppliedPatterns

  /**
   * Retrieves the flattening options so far.
   * @return the collection of flattening options
   *         (contains SolutionPatterns and/or null).
   */
  public Collection getOptions()
  {
    return options_;
  } //getOptions

  /**
   * Checks a Concept object for translation to the target language.
   * @param concept - the concept to check.
   * @throws CoComposeException - if the concept does not translate.
   */
  public void visitConcept(Concept concept) throws CoComposeException
  {
    String language = getTargetLanguage();
    Enumeration impls;
    Implementation impl;
//    boolean works = false;

    if (log != null)
    {
      log.log("Checking concept " + concept.getName() +
        " for translation to " + language);
      log.incLogLevel();
    }
    try
    {
      // if no implementations, translation of concept only depends on composites
      if (concept.getImplementationCount() > 0)
      { // otherwise check
        impls = concept.getImplementations(language);
        if (!impls.hasMoreElements())
//        {
          //impl = (Implementation) impls.nextElement();
          // if there is an implementation for the given language, concept can
          // translate
          //if (language.equals(impl.getLanguage()))
//          works = true;
//        } //while
//        if (!works)
        {
          if (log != null)
            log.log("No implementation works for " + concept.getName());
          throw new CoComposeException("No implementation works for " +
            concept.getName(), concept);
        }
      }
    }
    finally
    {
      if (log != null)
        log.decLogLevel();
    }
  } //visitConcept

  /**
   * Gets the flattening options (solution patterns or nothing) that can be used
   * for flattening the composite.
   * @param composite - the composite to visit.
   * @throws CoComposeException - if no flattening is possible.
   */
  public void visitComposite(Composite composite) throws CoComposeException
  {
    Enumeration solutions;
    SolutionPattern subSolution;
    String language = getTargetLanguage();

    if (log != null)
    {
      log.log("Determining flattening options for composite " + composite.getName());
      log.incLogLevel();
    }
    try
    {
      // if has implementation generators in the given language, add option null
      if ((composite.hasImplementations(language)) && (!(composite.getPublishedConcepts().hasMoreElements())))
      {
        if (log != null)
          log.log("There are implementation generators for " + language);
        addOption(null);
      }
      // if solution patterns exist, check
      if (composite.hasSolutions())
      {
        solutions = composite.getSolutions();
        while (solutions.hasMoreElements())
        {
          subSolution = (SolutionPattern) solutions.nextElement();
          try
          {
            // check for solution pattern loop
            if (!patternsInPath.contains(subSolution.getFilename()))
            {
              // check this solution for translation into the target language
              subSolution.accept(this);
              addOption(subSolution);
            }
          }
          catch (CoComposeException e)
          { // this solution didn't work, try next
            if (log != null)
              log.log(subSolution.getFilename() + " does not translate to " +
                language + " (" + e.getMessage() + ")");
          }
        } //while
      } //if
      if (getOptions().isEmpty())
      {
        if (log != null)
          log.log("No flattening options for the " + composite.getName() +
          " composite for translating to " + language);
        throw new CoComposeException("No flattening options for the " +
          composite.getName() + " composite for translating to " + language, composite);
      }
    }
    finally
    {
      if (log != null)
        log.decLogLevel();
    }
  } //visitComposite

  /**
   * Checks whether the solution pattern can eventually translate into the
   * given language.
   * @param solution - the solution pattern to visit.
   * @throws CoComposeException - if translation is not possible.
   */
  public void visitSolutionPattern(SolutionPattern solution) throws CoComposeException
  {
    Model tmpModel;
    Composite composite, thisComposite = solution.getComposite();
    Concept concept;
    File pathF;
    String path;
    FlatteningOptionsVisitor nestedOptions;
    Iterator concepts;
    Enumeration pubConcepts;
    AbstractConcept tmpconcept;

    if (log != null)
    {
      log.log("Checking solution pattern " + solution.getFilename() +
        " for translation to " + getTargetLanguage());
      log.incLogLevel();
    }
    try
    {
      pubConcepts = thisComposite.getPublishedConcepts();
      if (((appliedPatterns_ != null) && (!appliedPatterns_.isValidatedPattern(solution.getFilename()))) ||
          (pubConcepts.hasMoreElements()))
      { // check nested composites for solutions
        if (appliedPatterns_.containsPattern(solution.getFilename()))
        {
          tmpModel = appliedPatterns_.getSolutionModel(solution.getFilename());
        }
        else
        { // load solution
          pathF = new File(solution.getFilename());
          path = solution.getFilename();
          String name = pathF.getName();
          if (name.endsWith(".xml"))
            name = name.substring(0, name.length() - 4);
          try
          {
            tmpModel = new Model(path, name);
          }
          catch (Exception e)
          {
            if (log != null)
              log.log("Could not load solution pattern " + path + ": " +
                e.getMessage());
            throw new CoComposeException("Could not load solution pattern " + path +
              ": " + e.getMessage(), solution.getComposite());
          }
        }
        // check for published concepts
        while (pubConcepts.hasMoreElements())
        {
          tmpconcept = (PublishedConcept) pubConcepts.nextElement();
          if (tmpModel.getAbstractConcept(tmpconcept.getName(), tmpModel.getName()) == null)
          {
            throw new CoComposeException("Published concept " + tmpconcept.getName() + 
              " not found in " + solution.getFilename(), tmpconcept);
          }
        }
        if ((appliedPatterns_ != null) && (!appliedPatterns_.isValidatedPattern(solution.getFilename())))
        {
          // check concepts
          concepts = tmpModel.getAbstractConcepts().iterator();
          while (concepts.hasNext())
          {
            tmpconcept = (AbstractConcept) concepts.next();
            if (tmpconcept instanceof Concept)
              ((Concept) tmpconcept).accept(this);
          } //for
          // check composites
          concepts = tmpModel.getAbstractConcepts().iterator();
          while (concepts.hasNext())
          {
            tmpconcept = (AbstractConcept) concepts.next();
            if (tmpconcept instanceof Composite)
            {
              composite = (Composite) tmpconcept;
              try
              { // copy this visitor
                nestedOptions = (FlatteningOptionsVisitor) clone();
                nestedOptions.patternsInPath.add(solution.getFilename());
                // apply to nested composite
                composite.accept(nestedOptions);
              }
              catch (CloneNotSupportedException e)
              {
                if (log != null)
                  log.log("Could not clone visitor: " + e.getMessage());
                throw new CoComposeException("Could not clone visitor: " +
                  e.getMessage(), solution.getComposite());
              }
            }
          } //for
          // we made it, so cache this solution pattern
          appliedPatterns_.addValidatedPattern(solution);
        }
      } //if
    }
    finally
    {
      if (log != null)
        log.decLogLevel();
    }
  } //visitSolutionPattern
} //class FlatteningOptionsVisitor
