/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002, 2003
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package EDU.utwente.CoCompose.translate;

import java.io.*;
import java.util.*;
import org.xml.sax.XMLReader;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.helpers.AttributesImpl;
import EDU.utwente.CoCompose.lang.*;
import EDU.utwente.CoCompose.util.*;
import EDU.utwente.CoCompose.xml.ImplementationGeneratorReader;

/**
 * Represents an implementation generator for a Composite. It links to the files
 * containing the implementation generator declaration and the actual
 * implementation generator class.
 *
 * @author Dennis Wagelaar
 * @version 0.1.18, 22 November 2003
 */

public class ImplementationGenerator
{
  //////////////////////////////////////////////////////////////////////////////
  // Constants
  //////////////////////////////////////////////////////////////////////////////

  // XML parser defaults
  private static boolean setValidation        = false;
  private static boolean setNameSpaces        = true;
  private static boolean setSchemaSupport     = true;
  private static boolean setSchemaFullSupport = false;

  // XML System DTD path relative to application path
  private static final String XMLFILE_SYSTEM   = "dtd/ImplementationGenerator.dtd";
  // XML names
  private static final String XMLNAME_IMPLEMENTATION = "ImplementationGenerator";

  // Default parser name.
  private static final String
    DEFAULT_PARSER_NAME = "org.apache.xerces.parsers.SAXParser";

  //////////////////////////////////////////////////////////////////////////////
  // Attributes
  //////////////////////////////////////////////////////////////////////////////

  private String language_ = new String();
  private String filename_ = new String();
  private Composite composite_;
  private Vector roleparts_ = new Vector();
  private Hashtable rolepartsByRole_ = new Hashtable();

  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs an ImplementationGenerator.
   */
  public ImplementationGenerator()
  {
    super();
  } //ImplementationGenerator

  //////////////////////////////////////////////////////////////////////////////
  // Private methods
  //////////////////////////////////////////////////////////////////////////////

  /* Applies the solution pattern role part for the given concept. */
  private void applyRP(CompositeRole role, Concept concept) throws Exception
  {
    //stub
  } //applyRP

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Gets the language in which this solution pattern is expressed.
   *
   * @return The solution pattern language.
   */
  public String getLanguage()
  {
    return language_;
  } //getLanguage

  /**
   * Sets the language in which this solution pattern is expressed.
   *
   * @param language The solution pattern language.
   */
  public void setLanguage(String language)
  {
    language_ = language;
  } //setLanguage

  /**
   * Gets the name of the file that contains the actual solution pattern.
   *
   * @return The file name.
   */
  public String getFilename()
  {
    return filename_;
  } //getFilename

  /**
   * Sets the name of the file that contains the actual solution pattern.
   *
   * @param filename The file name.
   * @throws CoComposeException
   *   if the file specified by the filename could not be parsed.
   */
  public void setFilename(String filename) throws CoComposeException
  {
    // compensate for missing "./" at beginning (MacOS X)
    if (!filename.startsWith("."))
      filename = "./" + filename.replace('\\', '/');
    File file = new File(filename);
    filename_ = filename;
    // (re-)read XML data from file
    // create XML parser
    try
    {
      XMLReader xmlParser = (XMLReader)
        Class.forName(DEFAULT_PARSER_NAME).newInstance();
      xmlParser.setFeature( "http://xml.org/sax/features/validation",
        setValidation);
      xmlParser.setFeature( "http://xml.org/sax/features/namespaces",
        setNameSpaces);
      xmlParser.setFeature( "http://apache.org/xml/features/validation/schema",
        setSchemaSupport );
      xmlParser.setFeature( "http://apache.org/xml/features/validation/schema-full-checking",
        setSchemaFullSupport );
      // hook parser up to reader
      xmlParser.setContentHandler(new ImplementationGeneratorReader(this));
      // start parser
      xmlParser.parse(file.getAbsolutePath().replace('\\', '/'));
    }
    catch (Exception e)
    {
      throw new CoComposeException(e.getMessage());
    }
  } //setFilename

  /**
   * Gets the composite to which this solution pattern belongs.
   *
   * @return The composite that owns this solution pattern.
   */
  public Composite getComposite()
  {
    return composite_;
  } //getComposite

  /**
   * Sets the composite to which this solution pattern belongs.
   *
   * @param composite The composite that owns this solution pattern.
   */
  public void setComposite(Composite composite)
  {
    composite_ = composite;
  } //setComposite

  /**
   * Retrieves the model context.
   *
   * @return The model.
   */
  public Model getModel()
  {
    Model model = null;
    Composite composite = getComposite();
    if (composite != null)
      model = composite.getModel();
    return model;
  } //getModel

  /**
   * Gets the implementation generator role parts.
   *
   * @return An enumeration of ImplementationRolePart objects.
   */
  public Enumeration getRoleParts()
  {
    return roleparts_.elements();
  } //getRoleParts

  /**
   * Gets the implementation generator role parts for a given role name.
   * @param roleName the role name of the role parts.
   * @return An enumeration of ImplementationRolePart objects.
   */
  public Enumeration getRoleParts(String roleName)
  {
    Vector parts = (Vector) rolepartsByRole_.get(roleName);
    if (parts == null)
    {
      parts = new Vector();
    }
    return parts.elements();
  } //getRoleParts

  /**
   * Adds a role part to the implementation generator.
   *
   * @param rolepart The role part to add.
   */
  public void addRolePart(ImplementationRolePart rolepart)
  {
    roleparts_.add(rolepart);
    Vector parts = (Vector) rolepartsByRole_.get(rolepart.getRoleName());
    if (parts == null)
    {
      parts = new Vector();
      rolepartsByRole_.put(rolepart.getRoleName(), parts);
    }
    parts.add(rolepart);
    rolepart.setImplementationGenerator(this);
  } //addRolePart

  /**
   * Copies the attributes from the given implementation generator to self.
   * It doesn't copy the composite link.
   *
   * @param fromImplementation The implementation generator to copy from.
   * @throws CoComposeException if the copy could not be made.
   */
  public void copyAttributes(ImplementationGenerator fromImplementation)
    throws CoComposeException
  {
    setLanguage(fromImplementation.getLanguage());
    setFilename(fromImplementation.getFilename());
  } //copyAttributes

  /**
   * Gets the possible concept forms for a concept filling a role part of this
   * implementation generator.
   *
   * @param roleName The role name of the role to get the construct forms for.
   * @param logger The object to log to.
   * @return The set of possible concept forms for the given role.
   */
  public Set getRoleForms(String roleName, Logger logger)
  {
    ImplementationRolePart rolepart;
    HashSet result = new HashSet();

    if (logger != null)
    {
      logger.log("Retrieving concept forms for " + roleName +
        " role part of implementation generator " + getFilename());
      logger.incLogLevel();
    }
    try
    {
      Enumeration roleparts = getRoleParts();
      while (roleparts.hasMoreElements())
      { // check each role part
        rolepart = (ImplementationRolePart) roleparts.nextElement();
        if (rolepart.getRoleName().equals(roleName))
        { // role part corresponds with role name
          result.add(rolepart.getConceptForm());
          if (logger != null)
            logger.log("Adding concept form " + rolepart.getConceptForm());
        }
      }
    }
    finally
    {
      if (logger != null)
        logger.decLogLevel();
    }
    return result;
  } //getRoleForms

  /**
   * Create XML from this object using a SAX DefaultHandler.
   *
   * @param handler The SAX DefaultHandler.
   * @throws SAXException if an XML handler error occurs.
   */
  public void toXML(DefaultHandler handler) throws SAXException
  {
    // Create attributes
    AttributesImpl atts = new AttributesImpl();
    atts.addAttribute(null, null, "Language", "CDATA", getLanguage());
    atts.addAttribute(null, null, "Filename", "CDATA", getFilename());

    // Write XML
    handler.startElement(null, null, XMLNAME_IMPLEMENTATION, atts);
    handler.endElement(null, null, XMLNAME_IMPLEMENTATION);
  } //toXML

  /**
   * Accepts a ModelVisitor.
   *
   * @param visitor The visitor object to accept.
   * @throws CoComposeException if visiting operation failed.
   */
  public void accept(ModelVisitor visitor) throws CoComposeException
  {
    visitor.visitImplementationGenerator(this);
  } //accept
} //class ImplementationGenerator
