/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002, 2003
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package EDU.utwente.CoCompose.translate;

import java.util.*;
import EDU.utwente.CoCompose.lang.*;
import EDU.utwente.CoCompose.util.*;

/**
 * Represents a role part / concept form combination for an implementation
 * pattern.
 *
 * @author Dennis Wagelaar
 * @version 0.1.10, 22 July 2003
 */

public class ImplementationRolePart
{
  //////////////////////////////////////////////////////////////////////////////
  // Attributes
  //////////////////////////////////////////////////////////////////////////////

  private String rolename_;
  private String conceptform_;
  private Hashtable constraints_ = new Hashtable();
  private ImplementationGenerator implementation_;

  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs an ImplementationRolePart.
   */
  public ImplementationRolePart()
  {
    super();
  } //ImplementationRolePart

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  public ImplementationGenerator getImplementationGenerator()
  {
    return implementation_;
  } //getImplementationGenerator

  public void setImplementationGenerator(ImplementationGenerator implementation)
  {
    implementation_ = implementation;
  } //setImplementationGenerator

  /**
   * Gets the role name for this role part.
   *
   * @return The role name.
   */
  public String getRoleName()
  {
    return rolename_;
  } //getRoleName

  /**
   * Sets the role name for this role part.
   *
   * @param rolename The role name.
   */
  public void setRoleName(String rolename)
  {
    rolename_ = rolename;
  } //setRoleName

  /**
   * Gets the concept form for this role part.
   *
   * @return The concept form.
   */
  public String getConceptForm()
  {
    return conceptform_;
  } //getConceptForm

  /**
   * Sets the concept form for this role part.
   *
   * @param conceptform The concept form.
   */
  public void setConceptForm(String conceptform)
  {
    conceptform_ = conceptform;
  } //setConceptForm

  /**
   * Adds a constraint for this role part.
   *
   * @param rolename The name of the role upon which the constraint is based.
   * @param notEqualTo The concept form of the role indicated by the role name.
   */
  public void addConstraint(String rolename, String notEqualTo)
  {
    Vector notEqualVec = (Vector) constraints_.get(rolename);
    if (notEqualVec == null)
    {
      notEqualVec = new Vector();
      constraints_.put(rolename, notEqualVec);
    }
    notEqualVec.add(notEqualTo);
  } //addConstraint

  /**
   * Gets all constraints based upon the given role.
   *
   * @param rolename The name of the role on which the constraints are based.
   * @return The enumeration of required concept forms for the given role.
   */
  public Enumeration getConstraints(String rolename)
  {
    Vector notEqualVec = null;

    if (rolename != null)
    {
      notEqualVec = (Vector) constraints_.get(rolename);
    }
    if (notEqualVec == null)
    { // no constraint list defined for role
      notEqualVec = new Vector();
    }
    return notEqualVec.elements();
  } //getConstraints

  /**
   * Retrieves the role names for which constraints exist.
   *
   * @return An enumeration of role names.
   */
  public Enumeration getConstraintRoles()
  {
    return constraints_.keys();
  } //getConstraintRoles

  /**
   * Retrieves the composite the implementation generator implements.
   *
   * @return The composite.
   */
  public Composite getComposite()
  {
    Composite composite = null;
    ImplementationGenerator implementation = getImplementationGenerator();
    if (implementation != null)
      composite = implementation.getComposite();
    return composite;
  } //getComposite

  /**
   * Retrieves the model context.
   *
   * @return The model.
   */
  public Model getModel()
  {
    Model model = null;
    ImplementationGenerator implementation = getImplementationGenerator();
    if (implementation != null)
      model = implementation.getModel();
    return model;
  } //getModel

  /**
   * Accepts a ModelVisitor.
   *
   * @param visitor The visitor object to accept.
   * @throws CoComposeException if visiting operation failed.
   */
  public void accept(ModelVisitor visitor) throws CoComposeException
  {
    visitor.visitImplementationRolePart(this);
  } //accept
} //class ImplementationRolePart
