/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002, 2003
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package EDU.utwente.CoCompose.xml;

import java.io.*;
import java.util.*;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;
import EDU.auburn.VGJ.graph.*;
import EDU.auburn.VGJ.util.*;
import EDU.utwente.CoCompose.lang.*;
import EDU.utwente.CoCompose.translate.*;
import EDU.utwente.CoCompose.util.*;

/**
 * XML reader for reading a CoCompose Model XML document and create the
 * corresponding objects. Based on the SAX2 DefaultHandler class.
 *
 * @author Dennis Wagelaar
 * @version 0.1.13, 9 October 2003
 */

public class ModelReader extends DefaultHandler
{
  //////////////////////////////////////////////////////////////////////////////
  // Constants
  //////////////////////////////////////////////////////////////////////////////

  // XML Element names
  public static final String XMLNAME_MODEL     = "DesignModel";
  public static final String XMLNAME_SOLROLE   = "SolutionRole";
  public static final String XMLNAME_CONCEPT   = "Concept";
  public static final String XMLNAME_COMPOSITE = "Composite";
  public static final String XMLNAME_ROLE      = "CompositeRole";
  public static final String XMLNAME_PUBLISHED = "PublishedConcept";
  public static final String XMLNAME_LINK      = "ConceptLink";
  public static final String XMLNAME_GML       = "GML";
  public static final String XMLNAME_COMMENTS  = "Comments";
  public static final String XMLNAME_SPEC      = "Specification";
  public static final String XMLNAME_MULT      = "Multiplicity";
  public static final String XMLNAME_SOLPATT   = "SolutionPattern";
  public static final String XMLNAME_IMPLGEN   = "ImplementationGenerator";
  public static final String XMLNAME_IMPL      = "Implementation";
  public static final String XMLNAME_ICONT     = "ImplementationContent";
  public static final String XMLNAME_ICONSTR   = "ImplementationConstraint";
  public static final String XMLNAME_FORM      = "ConceptForm";
  public static final String XMLNAME_PRIORITY  = "TranslationPriority";
  public static final String XMLNAME_PRI_ITEM  = "PriorityItem";
  // Mode value
  public static final int CREATING_NONE      = 0;
  public static final int CREATING_SOLROLE   = 1;
  public static final int CREATING_CONCEPT   = 2;
  public static final int CREATING_COMPOSITE = 3;
  public static final int CREATING_ROLE      = 4;
  public static final int CREATING_PUBLISHED = 5;
  public static final int CREATING_LINK      = 6;
  // PCDATA mode value
  public static final int PCDATA_NONE      = 0;
  public static final int PCDATA_GML       = 1;
  public static final int PCDATA_COMMENTS  = 2;
  public static final int PCDATA_SPEC      = 3;

  //////////////////////////////////////////////////////////////////////////////
  // Attributes
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Mode flag for checking what element was started.
   */
  protected int mode_ = CREATING_NONE;
  /**
   * Mode flag for checking what type of character data is coming in.
   */
  protected int pcdataMode_ = PCDATA_NONE;
  /**
   * The model that is being read.
   */
  protected Model model;
  /**
   * New SolutionRole.
   */
  protected SolutionRole newSolutionRole_;
  /**
   * New Concept.
   */
  protected Concept newConcept_;
  /**
   * New Composite.
   */
  protected Composite newComposite_;
  /**
   * New CompositeRole.
   */
  protected CompositeRole newRole_;
  /**
   * New PublishedConcept.
   */
  protected PublishedConcept newPublished_;
  /**
   * New Link.
   */
  protected Link newLink_;
  /**
   * New concept implementation.
   */
  protected Implementation newImpl_;
  /**
   * New translation priority.
   */
  protected String transPrior_;
  /**
   * New implementation constraint concept ID.
   */
  protected String implConstrConceptID_;
  /**
   * New implementation constraint concept forms collection.
   */
  protected Collection implConstrForms_;
  private Node linkHead_;
  private Node linkTail_;
  private String specLang_;
  private StringBuffer charData_;
  private Hashtable realms_ = new Hashtable();
  private HashSet realmNames_ = new HashSet();
  // Implementation constraints tables for postponing the insertion of
  // implementation constraints (IDREFs to upcoming IDs problem).
  // Structure: Implementation |-> (<ConceptID> |-> Collection(<ConceptForm>))
  private Hashtable constraintsTable_ = new Hashtable();

  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a ModelReader.
   *
   * @param model_in The model object to build from the XML file.
   * @throws SAXException if the model is null.
   */
  public ModelReader(Model model_in) throws SAXException
  {
    if (model_in == null)
    {
      throw new SAXException("Model cannot be null");
    }
    model = model_in;
  } //ModelReader(Model model)

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Start DesignModel element.
   *
   * @param attrs Element attributes.
   * @throws Exception if the attributes could not be read.
   */
  protected void startModel(Attributes attrs) throws Exception
  {
    model.setTargetLanguage(attrs.getValue("TargetLanguage"));
  } //startModel

  /**
   * Start AbstractConcept element.
   *
   * @param concept The new AbstractConcept to build.
   * @param attrs Element attributes.
   * @throws Exception if the attributes could not be read.
   */
  protected void startAbstractConcept(AbstractConcept concept, Attributes attrs)
    throws Exception
  {
    String IdString = attrs.getValue("Id");
    if (IdString != null)
    {
      int Id = Integer.parseInt(IdString);
      concept.setId(Id);
    }
    model.insertNode(concept);
    String name = attrs.getValue("Name");
    if (name != null)
      concept.setName(name);
    String partOf = attrs.getValue("PartOf");
    if (partOf != null)
    {
      int partOfId = Integer.parseInt(partOf);
      concept.setPartOf((SolutionRole) model.getNodeFromId(partOfId));
    }
  } //startAbstractConcept

  /**
   * Start ConstrainedConcept element.
   *
   * @param concept The new ConstrainedConcept to build.
   * @param attrs Element attributes.
   * @throws Exception if the attributes could not be read.
   */
  protected void startConstrainedConcept(ConstrainedConcept concept, Attributes attrs)
    throws Exception
  {
    startAbstractConcept(concept, attrs);
    concept.setCompletenessConstraint(attrs.getValue("CompletenessConstraint"));
    concept.setCorrectnessConstraint(attrs.getValue("CorrectnessConstraint"));
  } //startConstrainedConcept

  /**
   * Start SolutionRole element.
   *
   * @param attrs Element attributes.
   * @throws Exception if the attributes could not be read.
   */
  protected void startSolutionRole(Attributes attrs) throws Exception
  {
    newSolutionRole_ = new SolutionRole();
    startAbstractConcept(newSolutionRole_, attrs);
    mode_ = CREATING_SOLROLE;
  } //startConcept

  /**
   * Start Concept element.
   *
   * @param attrs Element attributes.
   * @throws Exception if the attributes could not be read.
   */
  protected void startConcept(Attributes attrs) throws Exception
  {
    DefaultUniqueRealm realm;

    newConcept_ = new Concept();
    // From realm
    String from = attrs.getValue("From");
    if (from != null)
    {
      if (realmNames_.contains(from))
      { // realm already exists
        realm = (DefaultUniqueRealm) realms_.get(from);
      }
      else
      { // new realm needed
        realm = new DefaultUniqueRealm();
        realm.setName(from);
        realmNames_.add(from);
        realms_.put(from, realm);
      }
      newConcept_.setUniqueRealm(realm);
    }
    // Concept form
    String form = attrs.getValue("Form");
    if (form != null)
    {
      newConcept_.setForm(form);
    }
    startAbstractConcept(newConcept_, attrs);
    mode_ = CREATING_CONCEPT;
  } //startConcept

  /**
   * Start Composite element.
   *
   * @param attrs Element attributes.
   * @throws Exception if the attributes could not be read.
   */
  protected void startComposite(Attributes attrs) throws Exception
  {
    DefaultUniqueRealm realm;

    newComposite_ = new Composite();
    // From realm
    String from = attrs.getValue("From");
    if (from != null)
    {
      if (realmNames_.contains(from))
      { // realm already exists
        realm = (DefaultUniqueRealm) realms_.get(from);
      }
      else
      { // new realm needed
        realm = new DefaultUniqueRealm();
        realm.setName(from);
        realmNames_.add(from);
        realms_.put(from, realm);
      }
      newComposite_.setUniqueRealm(realm);
    }
    startConstrainedConcept(newComposite_, attrs);
    String type = attrs.getValue("Type");
    if (type != null)
      newComposite_.setType(type);
    String iconOnly = attrs.getValue("ShowIconOnly");
    newComposite_.setShowIconOnly(iconOnly.equals("Yes"));
    mode_ = CREATING_COMPOSITE;
  } //startComposite

  /**
   * Start CompositePartConcept element.
   *
   * @param concept The composite part concept.
   * @param attrs Element attributes.
   * @throws Exception if the attributes could not be read.
   */
  protected void startCompositePartConcept(CompositePartConcept concept,
    Attributes attrs) throws Exception
  {
    newComposite_.addPartConcept(concept);
    startConstrainedConcept(concept, attrs);
  } //startCompositePartConcept

  /**
   * Start CompositeRole element.
   *
   * @param attrs Element attributes.
   * @throws Exception if the attributes could not be read.
   */
  protected void startRole(Attributes attrs) throws Exception
  {
    newRole_ = new CompositeRole();
    startCompositePartConcept(newRole_, attrs);
    mode_ = CREATING_ROLE;
  } //startRole

  /**
   * Start PublishedConcept element.
   *
   * @param attrs Element attributes.
   * @throws Exception if the attributes could not be read.
   */
  protected void startPublished(Attributes attrs) throws Exception
  {
    newPublished_ = new PublishedConcept();
    startCompositePartConcept(newPublished_, attrs);
    String publicName = attrs.getValue("PublicName");
    if (publicName != null)
      newPublished_.setPublicName(publicName);
    mode_ = CREATING_PUBLISHED;
  } //startPublished

  /**
   * Start Multiplicity element.
   *
   * @param attrs Element attributes.
   * @throws Exception if the attributes could not be read.
   */
  protected void startMult(Attributes attrs) throws Exception
  {
    String low = attrs.getValue("LowerBound");
    String high = attrs.getValue("UpperBound");
    int lowBound = CompositeRole.MULT_ZERO;
    int highBound = CompositeRole.MULT_MANY;
    if (low.equals("1"))
      lowBound = CompositeRole.MULT_ONE;
    if (high.equals("1"))
      highBound = CompositeRole.MULT_ONE;
    newRole_.setMultLowerBound(lowBound);
    newRole_.setMultUpperBound(highBound);
  } //startMult

  /**
   * Start SolutionPattern element.
   *
   * @param attrs Element attributes.
   * @throws Exception if the attributes could not be read.
   */
  protected void startSolutionPattern(Attributes attrs) throws Exception
  {
    SolutionPattern solution = new SolutionPattern();
    solution.setFilename(attrs.getValue("Filename"));
    newComposite_.addSolution(solution);
  } //startSolutionPattern

  /**
   * Start ImplementationGenerator element.
   *
   * @param attrs Element attributes.
   * @throws Exception if the attributes could not be read.
   */
  protected void startImplementationGenerator(Attributes attrs) throws Exception
  {
    ImplementationGenerator implementation = new ImplementationGenerator();
    implementation.setLanguage(attrs.getValue("Language"));
    implementation.setFilename(attrs.getValue("Filename"));
    newComposite_.addImplementation(implementation);
  } //startImplementationGenerator

  /**
   * Start Implementation element.
   *
   * @param attrs Element attributes.
   * @throws Exception if the attributes could not be read.
   */
  protected void startImplementation(Attributes attrs) throws Exception
  {
    newImpl_ = new Implementation();
    newImpl_.setLanguage(attrs.getValue("Language"));
    newImpl_.setConceptForm(attrs.getValue("ConceptForm"));
  } //startImplementation

  /**
   * Start ImplementationContent element.
   *
   * @param attrs Element attributes.
   */
  protected void startImplementationContent(Attributes attrs)
  {
    // make new buffer to store implementation
    charData_ = new StringBuffer();
  } //startImplementationContent

  /**
   * Start ImplementationConstraint element.
   *
   * @param attrs Element attributes.
   * @throws Exception if the attributes could not be read.
   */
  protected void startImplementationConstraint(Attributes attrs)
    throws Exception
  {
    implConstrConceptID_ = attrs.getValue("Concept");
    implConstrForms_ = new Vector();
  } //startImplementationConstraint

  /**
   * Start ConceptForm element.
   *
   * @param attrs Element attributes.
   * @throws Exception if the attributes could not be read.
   */
  protected void startConceptForm(Attributes attrs) throws Exception
  {
    String form = attrs.getValue("Form");
    implConstrForms_.add(form);
  } //startImplConstraintForm

  /**
   * End ImplementationConstraint element.
   */
  protected void endImplementationConstraint()
  {
    Hashtable implConstraints = (Hashtable) constraintsTable_.get(newImpl_);
    if (implConstraints == null)
      implConstraints = new Hashtable();
    implConstraints.put(implConstrConceptID_, implConstrForms_);
    constraintsTable_.put(newImpl_, implConstraints);
  } //endImplementationConstraint

  /**
   * Start ConceptLink element.
   *
   * @param attrs Element attributes.
   * @throws Exception if the attributes could not be read.
   */
  protected void startLink(Attributes attrs) throws Exception
  {
    linkHead_ = model.getNodeFromId(
      Integer.parseInt(attrs.getValue("HeadConcept")));
    linkTail_ = model.getNodeFromId(
      Integer.parseInt(attrs.getValue("TailConcept")));
    // already create a link, in case no GML follows
    newLink_ = new Link(linkTail_, linkHead_, new DPoint3[0]);
    mode_ = CREATING_LINK;
  } //startLink

  /**
   * Start TranslationPriority element.
   *
   * @param attrs Element attributes.
   * @throws Exception if the attributes could not be read.
   */
  protected void startPriority(Attributes attrs) throws Exception
  {
    transPrior_ = attrs.getValue("TargetLanguage");
  } //startPriority

  /**
   * Start PriorityItem element.
   *
   * @param attrs Element attributes.
   * @throws Exception if the attributes could not be read.
   */
  protected void startPriorityItem(Attributes attrs) throws Exception
  {
    String form = attrs.getValue("ConceptForm");
    model.addTransPriority(transPrior_, form);
  } //startPriorityItem

  /**
   * Start GML element.
   *
   * @param attrs Element attributes.
   */
  protected void startGML(Attributes attrs)
  {
    charData_ = new StringBuffer();
    pcdataMode_ = PCDATA_GML;
  } //startGML

  /**
   * Start Comments element.
   *
   * @param attrs Element attributes.
   */
  protected void startComments(Attributes attrs)
  {
    charData_ = new StringBuffer();
    pcdataMode_ = PCDATA_COMMENTS;
  } //startComments

  /**
   * Start Specification element.
   *
   * @param attrs Element attributes.
   */
  protected void startSpec(Attributes attrs)
  {
    specLang_ = attrs.getValue("Language");
    charData_ = new StringBuffer();
    pcdataMode_ = PCDATA_SPEC;
  } //startSpec

  /**
   * Finish ImplementationContent element.
   */
  protected void endImplementationContent()
  {
    // set implementation content
    newImpl_.setContent(charData_.toString().trim());
  } //endImplementationContent

  /**
   * Finish Implementation element.
   */
  protected void endImplementation()
  {
    // insert new implementation
    newConcept_.addImplementation(newImpl_);
  } //endImplementation

  /**
   * Finish Model element - handles extra operations after parsing.
   *
   * @throws SAXException if an error occurs in post processing.
   */
  protected void endModel() throws SAXException
  {
    String conceptID;
    Collection forms;
    Implementation impl;
    AbstractConcept concept;
    Hashtable implConstraints;

    // insert all concept implementation constraints
    Enumeration impls = constraintsTable_.keys();
    while (impls.hasMoreElements())
    {
      impl = (Implementation) impls.nextElement();
      implConstraints = (Hashtable) constraintsTable_.get(impl);
      Enumeration constraints = implConstraints.keys();
      while (constraints.hasMoreElements())
      {
        conceptID = (String) constraints.nextElement();
        forms = (Collection) implConstraints.get(conceptID);
        // retrieve concept and add constraint
        concept = (AbstractConcept) model.getNodeFromId(Integer.parseInt(conceptID));
        if (concept == null)
        {
          throw new SAXException("Could not retrieve concept with ID " +
            conceptID + " when trying to add implementation constraint");
        }
        impl.addConstraint(concept, forms);
      }
    }
  } //endModel

  /**
   * Parses the GML data into the concept.
   *
   * @param concept The Concept to parse the GML data into.
   * @param gml The GML data.
   * @throws SAXException if the GML data could not be parsed.
   */
  protected void parseGML(AbstractConcept concept, String gml) throws SAXException
  {
    StringBufferInputStream stream = new StringBufferInputStream(gml);
    GMLlexer lexer = new GMLlexer(stream);
    try
    {
      GMLobject gmlObj = new GMLobject(lexer, null);
      GMLobject nodeGml = gmlObj.getGMLSubObject("node", GMLobject.GMLlist, false);
      Node newNode = new Node(nodeGml);
      concept.copyAttributes(newNode);
    }
    catch (Exception exc)
    {
      throw new SAXException(exc);
    }
  } //parseGML

  /**
   * Parses the GML data into a new link.
   *
   * @param gml The GML data.
   * @throws SAXException if the GML data could not be parsed.
   */
  protected void parseGML(String gml) throws SAXException
  {
    StringBufferInputStream stream = new StringBufferInputStream(gml);
    GMLlexer lexer = new GMLlexer(stream);
    try
    {
      GMLobject gmlObj = new GMLobject(lexer, null);
      GMLobject edgeGml = gmlObj.getGMLSubObject("edge", GMLobject.GMLlist, false);
      newLink_ = new Link(linkTail_, linkHead_, edgeGml);
    }
    catch (Exception exc)
    {
      throw new SAXException(exc);
    }
  } //parseGML

  /**
   * Parse character data into the given concept.
   *
   * @param concept The Concept to parse character data for.
   * @param data The character data.
   * @throws SAXException if the character data could not be parsed.
   */
  protected void conceptChar(AbstractConcept concept) throws SAXException
  {
    switch (pcdataMode_)
    {
      case PCDATA_GML:
        parseGML(concept, charData_.toString().trim());
        break;
      case PCDATA_COMMENTS:
        concept.setComments(charData_.toString().trim());
        break;
      case PCDATA_SPEC:
        concept.setSpecification(charData_.toString().trim());
        concept.setSpecificationLang(specLang_);
        break;
    }
  } //conceptChar

  /**
   * Parse character data into the given edge.
   *
   * @param data The character data.
   * @throws SAXException if the character data could not be parsed.
   */
  protected void edgeChar() throws SAXException
  {
    switch (pcdataMode_)
    {
      case PCDATA_GML:
        parseGML(charData_.toString());
        break;
    }
  } //edgeChar

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Start element.
   *
   * @param uri Ignored.
   * @param local Ignored.
   * @param raw Element name.
   * @param attrs Element attributes.
   * @throws SAXException if the new element could not be parsed.
   */
  public void startElement(String uri, String local, String raw,
    Attributes attrs) throws SAXException
  {
    try
    {
      // Model
      if (raw.equals(XMLNAME_MODEL))
        startModel(attrs);
      // new SolutionRole
      else if (raw.equals(XMLNAME_SOLROLE))
        startSolutionRole(attrs);
      // new Concept
      else if (raw.equals(XMLNAME_CONCEPT))
        startConcept(attrs);
      // new Composite
      else if (raw.equals(XMLNAME_COMPOSITE))
        startComposite(attrs);
      // new CompositeRole
      else if (raw.equals(XMLNAME_ROLE))
        startRole(attrs);
      // new PublishedConcept
      else if (raw.equals(XMLNAME_PUBLISHED))
        startPublished(attrs);
      // new Multiplicity
      else if (raw.equals(XMLNAME_MULT))
        startMult(attrs);
      // new SolutionPattern
      else if (raw.equals(XMLNAME_SOLPATT))
        startSolutionPattern(attrs);
      // new ImplementationGenerator
      else if (raw.equals(XMLNAME_IMPLGEN))
        startImplementationGenerator(attrs);
      // new Implementation
      else if (raw.equals(XMLNAME_IMPL))
        startImplementation(attrs);
      // new ImplementationContent
      else if (raw.equals(XMLNAME_ICONT))
        startImplementationContent(attrs);
      // new ImplementationConstraint
      else if (raw.equals(XMLNAME_ICONSTR))
        startImplementationConstraint(attrs);
      // new ConceptForm
      else if (raw.equals(XMLNAME_FORM))
        startConceptForm(attrs);
      // new Link
      else if (raw.equals(XMLNAME_LINK))
        startLink(attrs);
      // new TranslationPriority
      else if (raw.equals(XMLNAME_PRIORITY))
        startPriority(attrs);
      // new PriorityItem
      else if (raw.equals(XMLNAME_PRI_ITEM))
        startPriorityItem(attrs);
      // parse GML
      else if (raw.equals(XMLNAME_GML))
        startGML(attrs);
      // parse Comments
      else if (raw.equals(XMLNAME_COMMENTS))
        startComments(attrs);
      // parse Specification
      else if (raw.equals(XMLNAME_SPEC))
        startSpec(attrs);
    }
    catch (Exception e)
    {
      e.printStackTrace();
      throw new SAXException(e);
    }
  } //startElement

  /**
   * Characters.
   *
   * @param ch[] Character input.
   * @param start Starting index of characters.
   * @param length Length of characters.
   */
  public void characters(char ch[], int start, int length)
  {
    charData_.append(ch, start, length);
  } //characters

  /**
   * End element.
   *
   * @param uri Ignored.
   * @param local Ignored.
   * @param raw Element name.
   * @throws SAXException if the new element could not be parsed.
   */
  public void endElement(String uri, String local, String raw)
    throws SAXException
  {
    // new Node
    if ((raw.equals(XMLNAME_SOLROLE)) ||
        (raw.equals(XMLNAME_CONCEPT))  ||
        (raw.equals(XMLNAME_COMPOSITE)))
    {
      mode_ = CREATING_NONE;
    }
    // new CompositePartConcept
    else if ((raw.equals(XMLNAME_ROLE)) ||
             (raw.equals(XMLNAME_PUBLISHED)))
    { // embedded in Composite
      mode_ = CREATING_COMPOSITE;
    }
    // new Edge
    else if (raw.equals(XMLNAME_LINK))
    {
      try
      {
        model.insertLink(newLink_);
      }
      catch (CoComposeException exc)
      {
        throw new SAXException(exc);
      }
      mode_ = CREATING_NONE;
    }
    // new ImplementationContent
    else if (raw.equals(XMLNAME_ICONT))
    {
      endImplementationContent();
    }
    // new Implementation
    else if (raw.equals(XMLNAME_IMPL))
    {
      endImplementation();
    }
    // new ImplementationConstraint
    else if (raw.equals(XMLNAME_ICONSTR))
    {
      endImplementationConstraint();
    }
    // PCDATA
    else if ((raw.equals(XMLNAME_GML))      ||
             (raw.equals(XMLNAME_COMMENTS)) ||
             (raw.equals(XMLNAME_SPEC)))
    {
      // check what we've read and process that
      switch (mode_)
      {
        case CREATING_SOLROLE:
          conceptChar(newSolutionRole_);
          break;
        case CREATING_CONCEPT:
          conceptChar(newConcept_);
          break;
        case CREATING_COMPOSITE:
          conceptChar(newComposite_);
          break;
        case CREATING_ROLE:
          conceptChar(newRole_);
          break;
        case CREATING_PUBLISHED:
          conceptChar(newPublished_);
          break;
        case CREATING_LINK:
          edgeChar();
          break;
      }
      pcdataMode_ = PCDATA_NONE;
    }
    // Model element
    else if (raw.equals(XMLNAME_MODEL))
    {
      endModel();
    }
  } //endElement
} //class ModelReader
